/* Quicksort paralelo y secuencial.
 * Uso:
 *      % ./pquicksort <tamano arreglo> <semilla>
 *
 * Ejemplo:
 *      % ./pquicksort 100 1
 *
 * Este programa genera un arreglo a que corresponde a una permutacion aleatoria
 * de los enteros entre 0 y n-1, en donde n es el primer parametro del programa.
 * Luego hace una copia de a en el arreglo b.  Ordena el arreglo a usando
 * quicksort secuencial (funcion quicksort) y ordena el arreglo b usando
 * 2 procesos (funcion pquicksort).  Finalmente verifica que ambos arreglos
 * sean iguales.
 * Se puede variar el arreglo generado cambiando el valor de la semilla.  Por
 * ejemplo usando 2 en vez de 1.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "util.h"

void swap(int a[], int i, int j) {
  int aux= a[i];
  a[i]= a[j];
  a[j]= aux;
}

int particionar(int a[], int i, int j) {
  swap(a, i, (i+j)/2);
  int h= i;

  /*
    +--+-----------+--------+--------------+
    |  |///////////|\\\\\\\\|              |
    +--+-----------+--------+--------------+
    i              h        k              j
  */

  int k;
  for (k= i+1; k<=j; ++k)
    if (a[k]<a[i])
      swap(a, ++h, k);
  swap(a, i, h);
  return h;
}

void quicksort(int a[], int i, int j) {
  if (i<j) {
    int h= particionar(a, i, j);
    quicksort(a, i, h-1);
    quicksort(a, h+1, j);
  }
}

void pquicksort(int a[], int i, int j) {
  if (i<j) {
    int h= particionar(a, i, j);
    /* printf("h=%d a[h]=%d\n", h, a[h]); */
    int fd[2];
    pipe(fd);
    pid_t pid= fork();
    if (pid==0) {
      close(fd[0]);
      quicksort(a, i, h-1);
      write(fd[1], &a[i], (h-i)*sizeof(int));
      exit(0);
    }
    else {
      close(fd[1]);
      quicksort(a, h+1, j);
      leer(fd[0], (char*)&a[i], (h-i)*sizeof(int));
      close(fd[0]);
      waitpid(pid, NULL, 0);
    }
  }
}

int main(int argc, char **argv) {
  if (argc!=3) {
    fprintf(stderr, "uso: %s <nro. de elementos> <semilla>\n", argv[0]);
    exit(1);
  }
  int n= atoi(argv[1]);
  unsigned int seed= atoi(argv[2]);
  int *a= malloc(n*sizeof(int));
  int *b= malloc(n*sizeof(int));
  srand(seed);
  int i;
  for (i=0; i<n; i++)
    a[i]= i;
  for (i=n-1; i>=0; i--) {
    int k= rand()%(n-i);
    swap(a, i, k);
  }
  for (i=0; i<n; i++) {
    b[i]= a[i];
    if (i<20)
      printf("%d ", a[i]);
  }
  printf("%s\n", n>20 ? "..." : "");
  quicksort(a, 0, n-1);
  pquicksort(b, 0, n-1);
  for (i=0; i<n; i++) {
    if (a[i]!=b[i]) {
      fprintf(stderr, "resultados son distintos\n");
      exit(1);
    }
    if (i<20)
      printf("%d ", b[i]);
  }
  printf("%s\n", n>20 ? "..." : "");
  return 0;
}

